module helmet. 
import ListUtil.
import sig. 

type   write   A -> o.

write T :- 
  term_to_string T S,
  print S.    

% Meta-Interpreter %

type   init    o.

type   isin    o -> o -> o  -> o. 

isin   A (A | Rest) Rest. 
isin   A (B | R) (B | Rest) :-  isin A R Rest.
isin   A (R | S) (R1 | S )  :-  isin A R R1.  
isin   A (R | S) (R  | S1)  :-  isin A S S1.  
isin   A  A anti. 

type check o -> o -> o -> o.

check   anti A A.

check   A A anti.

check   A Goal RGoal :-
  isin  A Goal RGoal.
 
check  (A | Head) Goal RGoal :-
  isin  A Goal AGoal,
  check Head AGoal RGoal.

type  check_head o -> o -> o -> o -> o -> o -> o.

check_head Head Goal State Theta RGoal RTheta :-
  check Head  Goal  RGoal,
  check State Theta RTheta.

type fresh_vars o -> o -> o.

fresh_vars (pi D) F:-
 sigma x \ (fresh_vars (D x) F).
 
fresh_vars (A & B) A1:- 
  fresh_vars A A1.

fresh_vars (A & B) B1:- 
  fresh_vars B B1.

fresh_vars (Head --> Body) (Head --> Body).
fresh_vars (A >> B || C)   (A >> B || C). 
fresh_vars (A || B)        (A || B). 
  
type   select  o -> list o -> list o -> o. 

select Clause Bounded RBounded :-
  memb_and_rest QClause Bounded RBounded,
  fresh_vars    QClause Clause.
  
select (Head --> Body) Bounded Bounded :- (Head  --> Body).
select (A >> B || C) Bounded Bounded :- (A >> B || C).
select (A || B) Bounded Bounded :- (A || B).

type   myspy   string -> list A -> o.

type   spyon   o.

type   execute o -> list o -> o -> o. 
type   bc_over o ->  o -> list o -> list o -> o -> o.
type   b       list A -> o.

execute Goal Bounded (anti | Theta) :-
 myspy "Anti-Theta" (Goal::b(Bounded)::(anti | Theta)::nil),
 execute Goal Bounded Theta.

execute ((A | B) | C) Bounded Theta:-
 myspy "Associativity: " (((A | B) | C)::b(Bounded)::Theta::nil),
 execute (A | (B | C)) Bounded Theta.

execute all  Bounded Theta.

execute (A & B) Bounded Theta :-
  myspy "And-R" ((A & B)::b(Bounded)::Theta::nil), 
  execute A Bounded Theta, 
  execute B Bounded Theta. 

execute (A => B) Bounded Theta :-
  myspy "Augment" ((A => B)::b(Bounded)::Theta::nil), 
  A => (execute B Bounded Theta). 

execute (A -o B) Bounded Theta :- 
  myspy "Lolli-R" ((A -o B)::b(Bounded)::Theta::nil),
  execute B (A::Bounded) Theta. 

execute (pi D)  Bounded Theta :-
  myspy "Forall-R" ((pi D)::b(Bounded)::Theta::nil),
  (pi x \ (execute (D x) Bounded Theta)).

execute (all  | Goal) Bounded Theta. 

execute (anti | Goal) Bounded Theta :-
  myspy "Anti-R" ((anti | Goal)::b(Bounded)::Theta::nil),
  execute Goal Bounded Theta. 

execute ((A & B) | Goal) Bounded Theta :-
 % myspy "And-R" (((A & B)| Goal)::b(Bounded)::Theta::nil), 
  execute (A | Goal) Bounded Theta, 
  execute (B | Goal) Bounded Theta. 

%execute ((A @ B)  | Goal) Bounded Theta :-
%  myspy "Cond-R" (((A @ B)| Goal)::b(Bounded)::Theta::nil), 
%  execute (A | Goal) Bounded Theta, 
%  execute (B | Goal) Bounded Theta. 

execute ((A => B) | Goal) Bounded Theta :-
  myspy "Augment" (((A => B)| Goal)::b(Bounded)::Theta::nil), 
  A => (execute (B | Goal) Bounded Theta). 

execute ((A -o B) | Goal) Bounded Theta :- 
  myspy "Lolli-R" (((A -o B)| Goal)::b(Bounded)::Theta::nil),
  execute (B | Goal) (A::Bounded) Theta. 

execute ( (pi D) | Goal) Bounded Theta :-
  myspy "Forall-R" (((pi D)| Goal)::b(Bounded)::Theta::nil),
  (pi x \ (execute ((D x)| Goal) Bounded Theta)).

execute Goal Bounded Theta:-
   select  Clause Bounded RBounded, 
   bc_over Clause Goal Bounded RBounded Theta.  

type check_body o -> o -> o -> list o -> o.

check_body     (pi QBody)   RGoal RTheta RBounded:-
  pi x \ 
   (check_body (QBody x)    RGoal RTheta RBounded).

check_body (Body || NState) RGoal RTheta RBounded:-
  if (RGoal  = anti)      
     (then  (NGoal = Body))      
     (else (if (Body = anti) 
               (then (NGoal = RGoal)) 
               (else (NGoal = (Body | RGoal)))
            )
     ),                        
     if (RTheta = anti) 
     (then (NTheta = NState)) 
     (else (if (NState = anti)
               (then (NTheta = RTheta))
               (else (NTheta = (NState | RTheta)))
            )
     ),
  execute  NGoal RBounded NTheta.


bc_over (Cond >> Head || State --> QBody) Goal Bounded RBounded Theta :-  
  myspy "Backchaining" 
  (Goal::b(Bounded)::Theta::
         (Cond >> Head || State --> QBody)::nil),        
  check_head  Head Goal State Theta RGoal RTheta,
  Cond,
  check_body QBody RGoal RTheta RBounded.
  
bc_over (Head || State --> QBody) Goal Bounded RBounded Theta :-  
  myspy "Backchaining" 
  (Goal::b(Bounded)::Theta::
         (Head || State --> QBody)::nil),        
  check_head  Head Goal State Theta RGoal RTheta,
  check_body QBody RGoal RTheta RBounded.

bc_over (Cond >> Head || State) Goal Bounded RBounded Theta:-  
  myspy "Backchaining" (Goal::b(Bounded)::Theta::
                        (Cond >> Head || State)::nil),       
  check_head  Head Goal State Theta anti anti,
  Cond. 

bc_over (Head || State) Goal Bounded RBounded Theta:-  
  myspy "Backchaining" (Goal::b(Bounded)::Theta::(Head || State)::nil),       
  check_head  Head Goal State Theta anti anti. 

execute anti nil Theta :-
 print  "\nFinal State: ",   
 write  Theta,
 print "\n".

type ehhfsolve   o -> o.

ehhfsolve Goal :- 
  initial(Theta), 
  execute Goal nil Theta.
 
myspy Title (Goal::b(Bounded)::Theta::Rest):-
  spyon,
  !,
  if (Rest = (B::nil)) 
    (then (Msg = ((msg "Applied Clause: " B)::nil)))
    (else (Msg = nil)),
  generic_spy Title 
             ((msg "Current Goal: "  Goal)::
              (msg "Current Bounded Context: " Bounded)::
              (msg "Current State: " Theta)::Msg),
  read Dummy.
 
myspy Type (Goal::b(Bounded)::Theta::Rest):-
  spyon,
  print(Type),
  print(":\n"),
  print("Current Goal:"),
  write(Goal),print("\n"),
  print("Current Bounded Context:"),
  write(Bounded),print("\n"),
  print("Current State:"),
  write(Theta),print("\n"),
  if_then_else (Rest = (B::nil) ) 
    (print("\n Applied Clause: ") & write(B) & print("\n\n"))
    (true),     
  read Dummy.

myspy T X.
   













